@@ -70,7 +70,8 @@ class mixedfilemodewrapper(object):
         object.__setattr__(self, r'_lastop', 0)
 
     def __enter__(self):
-        return self._fp.__enter__()
+        self._fp.__enter__()
+        return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         self._fp.__exit__(exc_type, exc_val, exc_tb)
@@ -132,7 +133,10 @@ class fdproxy(object):
         self._fp = fp
 
     def __enter__(self):
-        return self._fp.__enter__()
+        self._fp.__enter__()
+        # Return this wrapper for the context manager so that the name is
+        # still available.
+        return self
 
     def __exit__(self, exc_type, exc_value, traceback):
         self._fp.__exit__(exc_type, exc_value, traceback)
@@ -244,7 +248,7 @@ def sshargs(sshcmd, host, user, port):
 def setflags(f, l, x):
     pass
 
-def copymode(src, dst, mode=None):
+def copymode(src, dst, mode=None, enforcewritable=False):
     pass
 
 def checkexec(path):
